"""Get machine learning predictions from geodata raster images (web package)"""
import os
from pathlib import Path

# not used here but contextily_tile is imported in samgis.io.tms2geotiff
from contextily import tile as contextily_tile

PROJECT_ROOT_FOLDER = Path(globals().get("__file__", "./_")).absolute().parent.parent
WORKDIR = os.getenv("WORKDIR", PROJECT_ROOT_FOLDER)
PROJECT_MODEL_FOLDER = Path(PROJECT_ROOT_FOLDER / "machine_learning_models")
MODEL_FOLDER = os.getenv("MODEL_FOLDER", PROJECT_MODEL_FOLDER)



import os

from numpy import ndarray, ma
from samgis_core import app_logger
from samgis_core.utilities.type_hints import TupleFloat
from xyzservices import TileProvider

from samgis_web.utilities.constants import (OUTPUT_CRS_STRING, DRIVER_RASTERIO_GTIFF, N_MAX_RETRIES, N_CONNECTION,
                                            N_WAIT, ZOOM_AUTO, BOOL_USE_CACHE)
from samgis_web.utilities.type_hints import tuple_ndarray_transform


bool_use_cache = int(os.getenv("BOOL_USE_CACHE", BOOL_USE_CACHE))
n_connection = int(os.getenv("N_CONNECTION", N_CONNECTION))
n_max_retries = int(os.getenv("N_MAX_RETRIES", N_MAX_RETRIES))
n_wait = int(os.getenv("N_WAIT", N_WAIT))
zoom_auto_string = os.getenv("ZOOM_AUTO", ZOOM_AUTO)


def download_extent(w: float, s: float, e: float, n: float, zoom: int or str = zoom_auto_string,
                    source: TileProvider or str = None,
                    wait: int = n_wait, max_retries: int = n_max_retries, n_connections: int = n_connection,
                    use_cache: bool = bool_use_cache) -> tuple_ndarray_transform:
    """
    Download, merge and crop a list of tiles into a single geo-referenced image or a raster geodata

    Args:
        w: West edge
        s: South edge
        e: East edge
        n: North edge
        zoom: Level of detail
        source: The tile source: web tile provider or path to local file. The web tile provider can be in the form of
            a :class:`xyzservices.TileProvider` object or a URL. The placeholders for the XYZ in the URL need to be
            `{x}`, `{y}`, `{z}`, respectively. For local file paths, the file is read with `rasterio` and all bands are
            loaded into the basemap. IMPORTANT: tiles are assumed to be in the Spherical Mercator projection
            (EPSG:3857), unless the `crs` keyword is specified.
        wait: if the tile API is rate-limited, the number of seconds to wait
            between a failed request and the next try
        max_retries: total number of rejected requests allowed before contextily will stop trying to fetch more tiles
            from a rate-limited API.
        n_connections: Number of connections for downloading tiles in parallel. Be careful not to overload the tile
            server and to check the tile provider's terms of use before increasing this value. E.g., OpenStreetMap has
            a max. value of 2 (https://operations.osmfoundation.org/policies/tiles/). If allowed to download in
            parallel, a recommended value for n_connections is 16, and should never be larger than 64.
        use_cache: If False, caching of the downloaded tiles will be disabled. This can be useful in resource
            constrained environments, especially when using n_connections > 1, or when a tile provider's terms of use
            don't allow caching.

    Returns:
        parsed request input

    """
    try:
        from samgis_web import contextily_tile
        from samgis_web.io_package.coordinates_pixel_conversion import _from4326_to3857

        app_logger.info(f"connection number:{n_connections}, type:{type(n_connections)}.")
        app_logger.info(f"zoom:{zoom}, type:{type(zoom)}.")
        app_logger.debug(f"download raster from source:{source} with bounding box w:{w}, s:{s}, e:{e}, n:{n}.")
        app_logger.debug(f"types w:{type(w)}, s:{type(s)}, e:{type(e)}, n:{type(n)}.")
        downloaded_raster, bbox_raster = contextily_tile.bounds2img(
            w, s, e, n, zoom=zoom, source=source, ll=True, wait=wait, max_retries=max_retries,
            n_connections=n_connections, use_cache=use_cache)
        xp0, yp0 = _from4326_to3857(n, e)
        xp1, yp1 = _from4326_to3857(s, w)
        cropped_image_ndarray, cropped_transform = crop_raster(yp1, xp1, yp0, xp0, downloaded_raster, bbox_raster)
        return cropped_image_ndarray, cropped_transform
    except Exception as e_download_extent:
        app_logger.exception(f"e_download_extent:{e_download_extent}.", exc_info=True)
        raise e_download_extent


def crop_raster(w: float, s: float, e: float, n: float, raster: ndarray, raster_bbox: TupleFloat,
                crs: str = OUTPUT_CRS_STRING, driver: str = DRIVER_RASTERIO_GTIFF) -> tuple_ndarray_transform:
    """
    Crop a raster using given bounding box (w, s, e, n) values

    Args:
        w: cropping west edge
        s: cropping south edge
        e: cropping east edge
        n: cropping north edge
        raster: raster image to crop
        raster_bbox: bounding box of raster to crop
        crs: The coordinate reference system. Required in 'w' or 'w+' modes, it is ignored in 'r' or 'r+' modes.
        driver: A short format driver name (e.g. "GTiff" or "JPEG") or a list of such names (see GDAL docs at
            https://gdal.org/drivers/raster/index.html ). In 'w' or 'w+' modes a single name is required. In 'r' or 'r+'
            modes the driver can usually be omitted. Registered drivers will be tried sequentially until a match is
            found. When multiple drivers are available for a format such as JPEG2000, one of them can be selected by
            using this keyword argument.

    Returns:
        cropped raster with its Affine transform

    """
    try:
        from rasterio.io import MemoryFile
        from rasterio.mask import mask as rio_mask
        from shapely.geometry import Polygon
        from geopandas import GeoSeries

        app_logger.debug(f"raster: type {type(raster)}, raster_ext:{type(raster_bbox)}, {raster_bbox}.")
        img_to_save, transform = get_transform_raster(raster, raster_bbox)
        img_height, img_width, number_bands = img_to_save.shape
        # https://rasterio.readthedocs.io/en/latest/topics/memory-files.html
        with MemoryFile() as rio_mem_file:
            app_logger.debug("writing raster in-memory to crop it with rasterio.mask.mask()")
            with rio_mem_file.open(
                    driver=driver,
                    height=img_height,
                    width=img_width,
                    count=number_bands,
                    dtype=str(img_to_save.dtype.name),
                    crs=crs,
                    transform=transform,
            ) as src_raster_rw:
                for band in range(number_bands):
                    src_raster_rw.write(img_to_save[:, :, band], band + 1)
            app_logger.debug("cropping raster in-memory with rasterio.mask.mask()")
            with rio_mem_file.open() as src_raster_ro:
                shapes_crop_polygon = Polygon([(n, e), (s, e), (s, w), (n, w), (n, e)])
                shapes_crop = GeoSeries([shapes_crop_polygon])
                app_logger.debug(f"cropping with polygon::{shapes_crop_polygon}.")
                cropped_image, cropped_transform = rio_mask(src_raster_ro, shapes=shapes_crop, crop=True)
                cropped_image_ndarray = reshape_as_image(cropped_image)
        app_logger.info(f"cropped image::{cropped_image_ndarray.shape}.")
        return cropped_image_ndarray, cropped_transform
    except Exception as e_crop_raster:
        try:
            app_logger.error(f"raster type:{type(raster)}.")
            app_logger.error(f"raster shape:{raster.shape}, dtype:{raster.dtype}.")
        except Exception as e_shape_dtype:
            app_logger.exception(f"raster shape or dtype not found:{e_shape_dtype}.", exc_info=True)
        app_logger.exception(f"e_crop_raster:{e_crop_raster}.", exc_info=True)
        raise e_crop_raster


def get_transform_raster(raster: ndarray, raster_bbox: TupleFloat) -> tuple_ndarray_transform:
    """
    Convert the input raster image to RGB and extract the Affine 

    Args:
        raster: raster image to geo-reference
        raster_bbox: bounding box of raster to crop

    Returns:
        rgb raster image and its Affine transform

    """
    try:
        from rasterio.transform import from_origin
        from numpy import array as np_array, linspace as np_linspace, uint8 as np_uint8
        from samgis_core.utilities.utilities import convert_ndarray_to_pil

        app_logger.debug(f"raster: type {type(raster)}, raster_ext:{type(raster_bbox)}, {raster_bbox}.")
        rgb = convert_ndarray_to_pil(raster.astype(np_uint8)).convert('RGB')
        np_rgb = np_array(rgb)
        img_height, img_width, _ = np_rgb.shape

        min_x, max_x, min_y, max_y = raster_bbox
        app_logger.debug(f"raster rgb shape:{np_rgb.shape}, raster rgb bbox {raster_bbox}.")
        x = np_linspace(min_x, max_x, img_width)
        y = np_linspace(min_y, max_y, img_height)
        res_x = (x[-1] - x[0]) / img_width
        res_y = (y[-1] - y[0]) / img_height
        transform = from_origin(x[0] - res_x / 2, y[-1] + res_y / 2, res_x, res_y)
        return np_rgb, transform
    except Exception as e_get_transform_raster:
        app_logger.error(f"arguments raster: {type(raster)}, {raster}.")
        app_logger.error(f"arguments raster_bbox: {type(raster_bbox)}, {raster_bbox}.")
        app_logger.exception(f"e_get_transform_raster:{e_get_transform_raster}.", exc_info=True)
        raise e_get_transform_raster


def reshape_as_image(arr: ndarray | ma.MaskedArray) -> ndarray:
    """
    Double swapping of image axis, needed in crop_raster().

    Args:
        arr: numpy raster array

    Returns:
        numpy raster array

    """
    try:
        from numpy import swapaxes

        return swapaxes(swapaxes(arr, 0, 2), 0, 1)
    except Exception as e_reshape_as_image:
        app_logger.error(f"arguments: {type(arr)}, {arr}.")
        app_logger.exception(f"e_reshape_as_image:{e_reshape_as_image}.", exc_info=True)
        raise e_reshape_as_image



"""Project constants"""
import os

INPUT_CRS_STRING = "EPSG:4326"
OUTPUT_CRS_STRING = "EPSG:3857"
DRIVER_RASTERIO_GTIFF = "GTiff"
ROOT = "/tmp"
CUSTOM_RESPONSE_MESSAGES = {
    200: "ok",
    400: "Bad Request",
    422: "Missing required parameter",
    500: "Internal server error"
}
TILE_SIZE = 256
EARTH_EQUATORIAL_RADIUS = 6378137.0
WKT_3857 = 'PROJCS["WGS 84 / Pseudo-Mercator",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,'
WKT_3857 += 'AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],'
WKT_3857 += 'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],'
WKT_3857 += 'PROJECTION["Mercator_1SP"],PARAMETER["central_meridian",0],PARAMETER["scale_factor",1],'
WKT_3857 += 'PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],'
WKT_3857 += 'AXIS["X",EAST],AXIS["Y",NORTH],EXTENSION["PROJ4","+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 '
WKT_3857 += '+x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext +no_defs"],AUTHORITY["EPSG","3857"]]'
SERVICE_NAME = "sam-gis"
DEFAULT_LOG_LEVEL = 'INFO'
RETRY_DOWNLOAD = 3
TIMEOUT_DOWNLOAD = 60
CALLBACK_INTERVAL_DOWNLOAD = 0.05
BOOL_USE_CACHE = True
N_WAIT = 0
N_MAX_RETRIES = 2
N_CONNECTION = 2
ZOOM_AUTO = "auto"
DEFAULT_URL_TILES = 'https://tile.openstreetmap.org/{z}/{x}/{y}.png'
DOMAIN_URL_TILES_MAPBOX = "api.mapbox.com"
RELATIVE_URL_TILES_MAPBOX = "v/mapbox.terrain-rgb/{zoom}/{x}/{y}{@2x}.pngraw?access_token={TOKEN}"
COMPLETE_URL_TILES_MAPBOX = f"https://{DOMAIN_URL_TILES_MAPBOX}/{RELATIVE_URL_TILES_MAPBOX}"
# https://s3.amazonaws.com/elevation-tiles-prod/terrarium/13/1308/3167.png
DOMAIN_URL_TILES_NEXTZEN = "s3.amazonaws.com"
RELATIVE_URL_TILES_NEXTZEN = "elevation-tiles-prod/terrarium/{z}/{x}/{y}.png"  # "terrarium/{z}/{x}/{y}.png"
COMPLETE_URL_TILES_NEXTZEN = f"https://{DOMAIN_URL_TILES_NEXTZEN}/{RELATIVE_URL_TILES_NEXTZEN}"
CHANNEL_EXAGGERATIONS_LIST = [2.5, 1.1, 2.0]
SLOPE_CELLSIZE = 61
MODEL_NAME = os.getenv("MODEL_NAME", "mobile_sam")
MSG_WRITE_TMP_ON_DISK = "found option to write images and geojson output: "
GRADIO_EXAMPLE_BODY_DICTLIST = {
    "bbox": {
        "ne": {"lat": 39.036252959636606, "lng": 15.040283203125002},
        "sw": {"lat": 38.302869955150044, "lng": 13.634033203125002}
    },
    "prompt": [{"type": "point", "data": {"lat": 38.48542007717153, "lng": 14.921846904165468}, "label": 0}],
    "zoom": 10, "source_type": "OpenStreetMap"
}
GRADIO_EXAMPLE_BODY_STRING_PROMPT = {
    "bbox": {
        "ne": {"lat": 46.17271333276639, "lng": 10.079505443573},
        "sw": {"lat": 46.1677724417049, "lng": 10.068830251693727}
    },
    "string_prompt": "",
    "zoom": 17,
    "source_type": "Esri.WorldImagery"
}
GRADIO_EXAMPLES_TEXT_LIST = [
    """You need to identify the areas with trees in this photogrammetric image. Please output segmentation mask.""",
    """You need to identify the areas with streets in this photogrammetric image. Please output segmentation mask.""",
    """You need to identify the houses in this photogrammetric image. Give me a segmentation mask and explain why.""",
    """Describe what do you see in this image.""",
]
GRADIO_MARKDOWN = """# [LISA](https://github.com/dvlab-research/LISA) + [SamGIS](https://github.com/trincadev/samgis-be) on Zero GPU!

This project aims to permit use of [LISA](https://github.com/dvlab-research/LISA) (Reasoning Segmentation via Large Language Model) applied to geospatial data thanks to [SamGIS](https://github.com/trincadev/samgis-be). In this space I adapted LISA to HuggingFace [lisa-on-cuda](https://huggingface.co/spaces/aletrn/lisa-on-cuda) ZeroGPU space.

This [home page project](https://huggingface.co/spaces/aletrn/samgis-lisa-on-zero) is a plane Gradio interface that take a json in input to translate it to a geojson. More information about these API implementation [here](
https://aletrn-samgis-lisa-on-zero.hf.space/docs). On this [blog page](https://trinca.tornidor.com/projects/lisa-adapted-for-samgis) you can find more details, including some request and response examples with the geojson map representations.

You can also find the alternative map interface [here](https://aletrn-samgis-lisa-on-zero.hf.space/lisa/) useful to create on the fly the payload requests and to represent the geojson response.
"""



"""functions useful to convert to/from latitude-longitude coordinates to pixel image coordinates"""
from samgis_core.utilities.type_hints import TupleFloat, TupleFloatAny

from samgis_core import app_logger
from samgis_web.utilities.constants import TILE_SIZE, EARTH_EQUATORIAL_RADIUS
from samgis_web.utilities.type_hints import ImagePixelCoordinates
from samgis_web.utilities.type_hints import LatLngDict


def _get_latlng2pixel_projection(latlng: LatLngDict) -> ImagePixelCoordinates:
    from math import log, pi, sin

    app_logger.debug(f"latlng: {type(latlng)}, value:{latlng}.")
    app_logger.debug(f'latlng lat: {type(latlng.lat)}, value:{latlng.lat}.')
    app_logger.debug(f'latlng lng: {type(latlng.lng)}, value:{latlng.lng}.')
    try:
        sin_y: float = sin(latlng.lat * pi / 180)
        app_logger.debug(f"sin_y, #1:{sin_y}.")
        sin_y = min(max(sin_y, -0.9999), 0.9999)
        app_logger.debug(f"sin_y, #2:{sin_y}.")
        x = TILE_SIZE * (0.5 + latlng.lng / 360)
        app_logger.debug(f"x:{x}.")
        y = TILE_SIZE * (0.5 - log((1 + sin_y) / (1 - sin_y)) / (4 * pi))
        app_logger.debug(f"y:{y}.")

        return {"x": x, "y": y}
    except Exception as e_get_latlng2pixel_projection:
        app_logger.error(f'args type:{type(latlng)}, {latlng}.')
        app_logger.exception(f'e_get_latlng2pixel_projection:{e_get_latlng2pixel_projection}.', exc_info=True)
        raise e_get_latlng2pixel_projection


def _get_point_latlng_to_pixel_coordinates(latlng: LatLngDict, zoom: int | float) -> ImagePixelCoordinates:
    from math import floor

    try:
        world_coordinate: ImagePixelCoordinates = _get_latlng2pixel_projection(latlng)
        app_logger.debug(f"world_coordinate:{world_coordinate}.")
        scale: int = pow(2, zoom)
        app_logger.debug(f"scale:{scale}.")
        return ImagePixelCoordinates(
            x=floor(world_coordinate["x"] * scale),
            y=floor(world_coordinate["y"] * scale)
        )
    except Exception as e_format_latlng_to_pixel_coordinates:
        app_logger.error(f'latlng type:{type(latlng)}, {latlng}.')
        app_logger.error(f'zoom type:{type(zoom)}, {zoom}.')
        app_logger.exception(f'e_format_latlng_to_pixel_coordinates:{e_format_latlng_to_pixel_coordinates}.',
                             exc_info=True)
        raise e_format_latlng_to_pixel_coordinates


def get_latlng_to_pixel_coordinates(
        latlng_origin_ne: LatLngDict,
        latlng_origin_sw: LatLngDict,
        latlng_current_point: LatLngDict,
        zoom: int | float,
        k: str
) -> ImagePixelCoordinates:
    """
    Parse the input request lambda event

    Args:
        latlng_origin_ne: NE latitude-longitude origin point
        latlng_origin_sw: SW latitude-longitude origin point
        latlng_current_point: latitude-longitude prompt point
        zoom: Level of detail
        k: prompt type

    Returns:
        ImagePixelCoordinates: pixel image coordinate point

    """
    app_logger.debug(f"latlng_origin - {k}: {type(latlng_origin_ne)}, value:{latlng_origin_ne}.")
    app_logger.debug(f"latlng_current_point - {k}: {type(latlng_current_point)}, value:{latlng_current_point}.")
    latlng_map_origin_ne = _get_point_latlng_to_pixel_coordinates(latlng_origin_ne, zoom)
    latlng_map_origin_sw = _get_point_latlng_to_pixel_coordinates(latlng_origin_sw, zoom)
    latlng_map_current_point = _get_point_latlng_to_pixel_coordinates(latlng_current_point, zoom)
    diff_coord_x = abs(latlng_map_origin_sw["x"] - latlng_map_current_point["x"])
    diff_coord_y = abs(latlng_map_origin_ne["y"] - latlng_map_current_point["y"])
    point = ImagePixelCoordinates(x=diff_coord_x, y=diff_coord_y)
    app_logger.debug(f"point type - {k}: {point}.")
    return point


def _from4326_to3857(lat: float, lon: float) -> TupleFloat or TupleFloatAny:
    from math import radians, log, tan

    x_tile: float = radians(lon) * EARTH_EQUATORIAL_RADIUS
    y_tile: float = log(tan(radians(45 + lat / 2.0))) * EARTH_EQUATORIAL_RADIUS
    return x_tile, y_tile


def _deg2num(lat: float, lon: float, zoom: int):
    from math import radians, pi, asinh, tan

    n = 2 ** zoom
    x_tile = ((lon + 180) / 360 * n)
    y_tile = (1 - asinh(tan(radians(lat))) / pi) * n / 2
    return x_tile, y_tile



import importlib.metadata


try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError or ImportError as e:
    print(f"metadata::e: {type(e)}, {e}: package installed?")
    __version__ = "1.0.0"
